/* AsyncHelper.cs
 * Provides a generic implementation asynchronous event invocation
 *
 * Distributed with no license what-so-ever - feel free to do
 * whatever the hell you want with this :-)  I would appreciate it
 * if you dropped me a line to let me know if you've found this
 * useful though.
 *
 * Have fun!
 *
 * Tomer Gabel
 * http://www.tomergabel.com
 * tomer@tomergabel.com
 *
 */

using System;
using System.Collections;
using System.Reflection;

namespace TomerGabel.Examples
{

	/// <summary>
	/// Helper class which provides generic asynchronous event invocation
	/// </summary>
	/// <remarks>
	/// This class is optimized for a build-once use-often event container; every instance of
	/// AsyncHelper is intended to use with a single event container instance. Static helper
	/// methods should make life easier when firing single-use events.
	/// 
	/// AsyncHelper should not be used in its current form for events that return a value or take
	/// ref/out parameters; it could be enhanced to support those features if necessary.
	/// </remarks>
	sealed public class AsyncHelper
	{
		/// <summary>
		/// Inner class; provides information about a particular event instance
		/// </summary>
		private class DelegateInfo
		{
			/// <summary>
			/// Handler (delegate) type for this event
			/// </summary>
			private Type m_handler;
			/// <summary>
			/// Event instance for this event
			/// </summary>
			private object m_eventInstance;

			/// <summary>
			/// Handler (delegate) type for this event
			/// </summary>
			public Type Handler
			{
				get { return m_handler; }
			}

			/// <summary>
			/// Event instance for this event
			/// </summary>
			public object EventInstance
			{
				get { return m_eventInstance; }
			}
			
			
			/// <summary>
			/// Creates and initializes an instance of DelegateInfo
			/// </summary>
			/// <param name="target">Target event container</param>
			/// <param name="eventName">Event name</param>
			public DelegateInfo( object target, string eventName )
			{
				m_eventInstance = target.GetType().GetField( eventName, BindingFlags.Instance | BindingFlags.NonPublic ).GetValue( target );
				m_handler       = target.GetType().GetEvent( eventName, BindingFlags.Instance | BindingFlags.Public    ).EventHandlerType;
			}
		}
		
		/// <summary>
		/// Event definition cache
		/// </summary>
		private Hashtable m_events = new Hashtable();
		/// <summary>
		/// Event container instance
		/// </summary>
		private object m_target;
		
		/// <summary>
		/// Creates and initializes an instance of AsyncHelper
		/// </summary>
		/// <param name="target">Event container instance</param>
		public AsyncHelper( object target )
		{
			m_target = target;
		}
		
		
		/// <summary>
		/// Fires an event on the prespecified event container
		/// </summary>
		/// <param name="eventName">Event name</param>
		/// <param name="para">Event parameters</param>
		public void Fire( string eventName, params object[] para )
		{
			DelegateInfo info;
			lock ( m_events.SyncRoot )
				if ( null == ( info = m_events[ eventName ] as DelegateInfo ) )
					m_events.Add( eventName, info = new DelegateInfo( m_target, eventName ) );

			// Generate parameter list
			ArrayList paraList = new ArrayList( para );
			paraList.Add( new AsyncCallback( asyncHandler ) );
			paraList.Add( info );
			
			info.Handler.InvokeMember(
				"BeginInvoke",
				BindingFlags.Public | BindingFlags.Instance | BindingFlags.InvokeMethod,
				null,
				info.EventInstance,
				paraList.ToArray()
				);
		}
		
		/// <summary>
		/// Fires an event on any event container. Does not invoke any user-defined callback.
		/// </summary>
		/// <param name="target">Event container instance</param>
		/// <param name="eventName">Event name</param>
		/// <param name="para">Event parameters</param>
		public static void Fire( object target, string eventName, params object[] para )
		{
			AsyncHelper helper = new AsyncHelper( target );
			helper.Fire( eventName, para );
		}
		/// <summary>
		/// Fires an event on any event container. After invocation completes, a user-defined
		/// callback is called.
		/// </summary>
		/// <param name="target">Event container instance</param>
		/// <param name="eventName">Event name</param>
		/// <param name="callback">User-defined callback; called when execution completes</param>
		/// <param name="para">Event parameters</param>
		public static void Fire( object target, string eventName, EventHandler callback, params object[] para )
		{
			AsyncHelper helper = new AsyncHelper( target );
			helper.Done += callback;
			helper.Fire( eventName, para );
		}

		/// <summary>
		/// User-defined callback; called when execution completes
		/// </summary>
		public event EventHandler Done;

		/// <summary>
		/// Handles the asynchoronous operation callback
		/// </summary>
		/// <param name="ar">Object implementing IAsyncResult</param>
		public void asyncHandler( IAsyncResult ar )
		{
			DelegateInfo info = (DelegateInfo) ar.AsyncState;
		
			// Invoke delegate's EndInvoke method (as per .NET 1.1 guidelines)
			info.Handler.InvokeMember(
				"EndInvoke",
				BindingFlags.Public | BindingFlags.Instance | BindingFlags.InvokeMethod,
				null,
				info.EventInstance,
				new object[] { ar }
				);
			
			if ( null != Done )
				Done( this, new EventArgs() );
		}
	}
}
